@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -22,12 +22,46 @@
 # This file ensures that renpy packages will be imported in the right
 # order.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+
+# All imports should go below renpy.compat.
+
+
+def update_path():
+    """
+    Update the __path__ of package, to import binary modules from a libexec
+    directory.
+    """
+    import sys
+    import os.path
+
+    name = sys._getframe(1).f_globals["__name__"]
+    package = sys.modules[name]
+    name = name.split(".")
+
+    try:
+        import _renpy
+        if hasattr(_renpy, '__file__'): # .so/.dll
+            libexec = os.path.dirname(_renpy.__file__)
+            package.__path__.append(os.path.join(libexec, *name))
+
+        # Also find encodings, to deal with the way py2exe lays things out.
+        import encodings
+        libexec = os.path.dirname(encodings.__path__[0])
+        package.__path__.append(os.path.join(libexec, *name))
+    except ImportError:
+        return
+
+
+from renpy.compat import *
+update_path()
+
+import renpy.compat.pickle as pickle
+
 import sys
 import os
 import copy
 import types
-import cPickle
 
 ################################################################################
 # Version information
@@ -35,17 +69,15 @@ import cPickle
 
 # Version numbers.
 try:
-    from renpy.vc_version import vc_version; vc_version
+    from renpy.vc_version import vc_version
 except ImportError:
     vc_version = 0
 
 # The tuple giving the version number.
-version_tuple = (7, 2, 0, vc_version)
+version_tuple = (7, 4, 4, vc_version)
 
 # The name of this version.
-version_name = "What's on the menu."
-# 7.3: The world (wide web) is not enough.
-
+version_name = "The Pot of Gold."
 
 # A string giving the version number only (8.0.1.123).
 version_only = ".".join(str(i) for i in version_tuple)
@@ -58,7 +90,6 @@ script_version = 5003000
 savegame_suffix = "-LT1.save"
 bytecode_version = 1
 
-
 ################################################################################
 # Platform Information
 ################################################################################
@@ -92,7 +123,7 @@ def get_windows_version():
                     ('dwMinorVersion', ctypes.c_ulong),
                     ('dwBuildNumber', ctypes.c_ulong),
                     ('dwPlatformId', ctypes.c_ulong),
-                    ('szCSDVersion', ctypes.c_wchar*128),
+                    ('szCSDVersion', ctypes.c_wchar * 128),
                     ('wServicePackMajor', ctypes.c_ushort),
                     ('wServicePackMinor', ctypes.c_ushort),
                     ('wSuiteMask', ctypes.c_ushort),
@@ -117,7 +148,7 @@ def get_windows_version():
 
 if platform.win32_ver()[0]:
     windows = get_windows_version()
-elif "RENPY_IOS" in os.environ:
+elif os.environ.get("RENPY_PLATFORM", "").startswith("ios"):
     ios = True
 elif platform.mac_ver()[0]:
     macintosh = True
@@ -134,7 +165,6 @@ mobile = android or ios or emscripten
 # A flag that's set to true if the game directory is bundled inside a mac app.
 macapp = False
 
-
 ################################################################################
 # Backup Data for Reload
 ################################################################################
@@ -146,7 +176,6 @@ safe_mode_checked = False
 # needs to survive through an utter restart.
 autoreload = False
 
-
 # A dict that persists through utter restarts. Accessible to all code as
 # renpy.session.
 session = { }
@@ -155,6 +184,8 @@ session = { }
 # to backup.
 backup_blacklist = {
     "renpy",
+    "renpy.compat",
+    "renpy.compat.dictviews",
     "renpy.object",
     "renpy.log",
     "renpy.bootstrap",
@@ -172,6 +203,8 @@ backup_blacklist = {
     "renpy.test.testkey",
     "renpy.test.testmouse",
     "renpy.test.testparser",
+    "renpy.gl2",
+    "renpy.gl",
     "renpycoverage",
     }
 
@@ -197,10 +230,13 @@ name_blacklist = {
     "renpy.display.screen.cprof",
     "renpy.audio.audio.lock",
     "renpy.audio.audio.periodic_condition",
+    "renpy.webloader.queue_lock",
+    "renpy.persistent.save_MP_instances",
+    "renpy.exports.sdl_dll",
     }
 
 
-class Backup():
+class Backup(object):
     """
     This represents a backup of all of the fields in the python modules
     comprising Ren'Py, shortly after they were imported.
@@ -232,7 +268,7 @@ class Backup():
             self.backup_module(m)
 
         # A pickled version of self.objects.
-        self.objects_pickle = cPickle.dumps(self.objects, cPickle.HIGHEST_PROTOCOL)
+        self.objects_pickle = pickle.dumps(self.objects, pickle.HIGHEST_PROTOCOL)
 
         self.objects = None
 
@@ -257,7 +293,7 @@ class Backup():
 
         self.names[mod] = set(vars(mod).keys())
 
-        for k, v in vars(mod).iteritems():
+        for k, v in vars(mod).items():
 
             if k.startswith("__") and k.endswith("__"):
                 continue
@@ -276,10 +312,10 @@ class Backup():
             # If we have a problem pickling things, uncomment the next block.
 
             try:
-                cPickle.dumps(v, cPickle.HIGHEST_PROTOCOL)
+                pickle.dumps(v, pickle.HIGHEST_PROTOCOL)
             except:
                 print("Cannot pickle", name + "." + k, "=", repr(v))
-                print("Reduce Ex is:", repr(v.__reduce_ex__(cPickle.HIGHEST_PROTOCOL)))
+                print("Reduce Ex is:", repr(v.__reduce_ex__(pickle.HIGHEST_PROTOCOL)))
 
     def restore(self):
         """
@@ -291,14 +327,14 @@ class Backup():
             return
 
         # Remove new variables from the module.
-        for mod, names in self.names.iteritems():
+        for mod, names in self.names.items():
             modvars = vars(mod)
             for name in set(modvars.keys()) - names:
                 del modvars[name]
 
-        objects = cPickle.loads(self.objects_pickle)
+        objects = pickle.loads(self.objects_pickle)
 
-        for k, v in self.variables.iteritems():
+        for k, v in self.variables.items():
             mod, field = k
             setattr(mod, field, objects[v])
 
@@ -311,28 +347,12 @@ backup = None
 ################################################################################
 
 
-def update_path(package):
-    """
-    Update the __path__ of package, to import binary modules from a libexec
-    directory.
-    """
-
-    name = package.__name__.split(".")
-
-    import _renpy
-    if hasattr(_renpy, '__file__'):  # .so/.dll
-        libexec = os.path.dirname(_renpy.__file__)
-        package.__path__.append(os.path.join(libexec, *name))
-
-    # Also find encodings, to deal with the way py2exe lays things out.
-    import encodings
-    libexec = os.path.dirname(encodings.__path__[0])
-    package.__path__.append(os.path.join(libexec, *name))
-
-# Replaced below.
-
-
 def plog(level, even, *args):
+    """
+    Empty version of renpy.plog that is replaced by the real implementation
+    in import_all.
+    """
+
     return
 
 
@@ -343,11 +363,12 @@ def import_all():
 
     # Note: If we add a new module, we need to add it to iOS.
 
-    import renpy  # @UnresolvedImport
+    # Note: If we add a new module, it should be added at the bottom of this file so it shows up in
+    # code analysis.
 
-    update_path(renpy)
+    import renpy # @UnresolvedImport
 
-    import renpy.arguments  # @UnresolvedImport
+    import renpy.arguments # @UnresolvedImport
 
     import renpy.config
     import renpy.log
@@ -374,7 +395,7 @@ def import_all():
     import renpy.easy
     import renpy.execution
     import renpy.loadsave
-    import renpy.savelocation  # @UnresolvedImport
+    import renpy.savelocation # @UnresolvedImport
     import renpy.persistent
     import renpy.scriptedit
     import renpy.parser
@@ -387,13 +408,12 @@ def import_all():
     global plog
     plog = renpy.performance.log
 
-    import renpy.styledata  # @UnresolvedImport
-    update_path(renpy.styledata)
+    import renpy.styledata # @UnresolvedImport
 
     import renpy.style
     renpy.styledata.import_style_functions()
 
-    sys.modules['renpy.styleclass'] = renpy.style
+    sys.modules[pystr('renpy.styleclass')] = renpy.style
 
     import renpy.substitutions
     import renpy.translation
@@ -403,21 +423,17 @@ def import_all():
     import renpy.translation.extract
     import renpy.translation.merge
 
-    import renpy.display  # @UnresolvedImport @Reimport
-
-    update_path(renpy.display)
+    import renpy.display # @UnresolvedImport @Reimport
 
     import renpy.display.presplash
     import renpy.display.pgrender
     import renpy.display.scale
     import renpy.display.module
-    import renpy.display.render  # Most display stuff depends on this. @UnresolvedImport
-    import renpy.display.core  # object @UnresolvedImport
+    import renpy.display.render # Most display stuff depends on this. @UnresolvedImport
+    import renpy.display.core # object @UnresolvedImport
 
     import renpy.text
 
-    update_path(renpy.text)
-
     import renpy.text.ftfont
     import renpy.text.font
     import renpy.text.textsupport
@@ -425,27 +441,22 @@ def import_all():
     import renpy.text.text
     import renpy.text.extras
 
-    sys.modules['renpy.display.text'] = renpy.text.text
+    sys.modules[pystr('renpy.display.text')] = renpy.text.text
 
     import renpy.gl
-    update_path(renpy.gl)
-
     import renpy.gl2
-    update_path(renpy.gl2)
-
     import renpy.angle
-    update_path(renpy.angle)
 
     import renpy.display.layout
     import renpy.display.viewport
     import renpy.display.transform
-    import renpy.display.motion  # layout @UnresolvedImport
-    import renpy.display.behavior  # layout @UnresolvedImport
-    import renpy.display.transition  # core, layout @UnresolvedImport
-    import renpy.display.movetransition  # core @UnresolvedImport
+    import renpy.display.motion # layout @UnresolvedImport
+    import renpy.display.behavior # layout @UnresolvedImport
+    import renpy.display.transition # core, layout @UnresolvedImport
+    import renpy.display.movetransition # core @UnresolvedImport
     import renpy.display.im
     import renpy.display.imagelike
-    import renpy.display.image  # core, behavior, im, imagelike @UnresolvedImport
+    import renpy.display.image # core, behavior, im, imagelike @UnresolvedImport
     import renpy.display.video
     import renpy.display.focus
     import renpy.display.anim
@@ -467,7 +478,6 @@ def import_all():
     # renpy.display.module.
 
     import renpy.audio
-    update_path(renpy.audio)
 
     import renpy.audio.audio
     import renpy.audio.music
@@ -477,8 +487,6 @@ def import_all():
     import renpy.screenlang
 
     import renpy.sl2
-    update_path(renpy.sl2)
-
     import renpy.sl2.slast
     import renpy.sl2.slparser
     import renpy.sl2.slproperties
@@ -492,13 +500,21 @@ def import_all():
     import renpy.memory
 
     import renpy.exports
-    import renpy.character  # depends on exports. @UnresolvedImport
+    import renpy.character # depends on exports. @UnresolvedImport
 
     import renpy.add_from
     import renpy.dump
 
-    import renpy.minstore  # depends on lots. @UnresolvedImport
-    import renpy.defaultstore  # depends on everything. @UnresolvedImport
+    import renpy.gl2.gl2draw
+    import renpy.gl2.gl2mesh
+    import renpy.gl2.gl2model
+    import renpy.gl2.gl2polygon
+    import renpy.gl2.gl2shader
+    import renpy.gl2.gl2texture
+    import renpy.gl2.live2d
+
+    import renpy.minstore # depends on lots. @UnresolvedImport
+    import renpy.defaultstore # depends on everything. @UnresolvedImport
 
     import renpy.test
     import renpy.test.testmouse
@@ -526,7 +542,7 @@ def post_import():
     of various modules.
     """
 
-    import renpy  # @UnresolvedImport
+    import renpy # @UnresolvedImport
 
     # Create the store.
     renpy.python.create_store("store")
@@ -538,19 +554,23 @@ def post_import():
     sys.modules['renpy.store'] = sys.modules['store']
 
     import subprocess
-    sys.modules['renpy.subprocess'] = subprocess
+    sys.modules[pystr('renpy.subprocess')] = subprocess
 
-    for k, v in renpy.defaultstore.__dict__.iteritems():
+    for k, v in renpy.defaultstore.__dict__.items():
         renpy.store.__dict__.setdefault(k, v)
 
     renpy.store.eval = renpy.defaultstore.eval
 
     # Import everything into renpy.exports, provided it isn't
     # already there.
-    for k, v in globals().iteritems():
+    for k, v in globals().items():
         vars(renpy.exports).setdefault(k, v)
 
 
+def issubmodule(sub, module):
+    return sub == module or sub.startswith(module + ".")
+
+
 def reload_all():
     """
     Resets all modules to the state they were in right after import_all
@@ -567,7 +587,7 @@ def reload_all():
     sys.exc_clear()
 
     # Reset the styles.
-    renpy.style.reset()  # @UndefinedVariable
+    renpy.style.reset() # @UndefinedVariable
 
     # Shut down the cache thread.
     renpy.display.im.cache.quit()
@@ -583,15 +603,16 @@ def reload_all():
     renpy.display.render.mark_sweep()
 
     # Get rid of the draw module and interface.
-    renpy.display.draw.deinit()
+    renpy.display.draw.quit()
     renpy.display.draw = None
     renpy.display.interface = None
 
     py_compile_cache = renpy.python.py_compile_cache
+    reload_modules = renpy.config.reload_modules
 
     # Delete the store modules.
-    for i in sys.modules.keys():
-        if i.startswith("store") or i == "renpy.store":
+    for i in list(sys.modules.keys()):
+        if issubmodule(i, "store") or i == "renpy.store":
             m = sys.modules[i]
 
             if m is not None:
@@ -599,6 +620,14 @@ def reload_all():
 
             del sys.modules[i]
 
+        elif any(issubmodule(i, m) for m in reload_modules):
+            m = sys.modules[i]
+
+            if m is not None:
+                m.__dict__.clear()
+
+            del sys.modules[i]
+
     # Restore the state of all modules from backup.
     backup.restore()
 
@@ -611,11 +640,11 @@ def reload_all():
     # Re-initialize the importer.
     renpy.loader.init_importer()
 
-
 ################################################################################
 # Fix things for code analysis
 ################################################################################
 
+
 def setup_modulefinder(modulefinder):
     """
     Informs modulefinder about the location of modules in nonstandard places.
@@ -625,7 +654,7 @@ def setup_modulefinder(modulefinder):
 
     libexec = os.path.dirname(_renpy.__file__)
 
-    for i in [ "display", "gl", "angle", "text", "styledata" ]:
+    for i in [ "compat", "display", "gl", "gl2", "angle", "text", "styledata" ]:
 
         displaypath = os.path.join(libexec, "renpy", i)
 
@@ -633,34 +662,165 @@ def setup_modulefinder(modulefinder):
             modulefinder.AddPackagePath('renpy.' + i, displaypath)
 
 
-def import_cython():
-    """
-    Never called, but necessary to ensure that modulefinder will properly
-    grab the various cython modules.
-    """
+# This is here to help code analysis tools figure out all the modules that get imported.
+# noinspection PyUnreachableCode
+if False:
+
+    import renpy # @UnresolvedImport
+
+    update_path(renpy)
+
+    import renpy.arguments # @UnresolvedImport
+
+    import renpy.config
+    import renpy.log
+
+    import renpy.display
+
+    import renpy.debug
+
+    # Should probably be early, as we will add it as a base to serialized things.
+    import renpy.object
+
+    import renpy.game
+    import renpy.preferences
+
+    # Adds in the Ren'Py loader.
+    import renpy.loader
+
+    import renpy.pyanalysis
+
+    import renpy.ast
+    import renpy.atl
+    import renpy.curry
+    import renpy.color
+    import renpy.easy
+    import renpy.execution
+    import renpy.loadsave
+    import renpy.savelocation # @UnresolvedImport
+    import renpy.persistent
+    import renpy.scriptedit
+    import renpy.parser
+    import renpy.performance
+    import renpy.pydict
+    import renpy.python
+    import renpy.script
+    import renpy.statements
+
+    import renpy.styledata # @UnresolvedImport
+    import renpy.style
+    import renpy.substitutions
+    import renpy.translation
+    import renpy.translation.scanstrings
+    import renpy.translation.generation
+    import renpy.translation.dialogue
+    import renpy.translation.extract
+    import renpy.translation.merge
+
+    import renpy.display # @UnresolvedImport @Reimport
+
+    import renpy.display.presplash
+    import renpy.display.pgrender
+    import renpy.display.scale
+    import renpy.display.module
+    import renpy.display.render # Most display stuff depends on this. @UnresolvedImport
+    import renpy.display.core # object @UnresolvedImport
+
+    import renpy.text
+    import renpy.text.ftfont
+    import renpy.text.font
+    import renpy.text.textsupport
+    import renpy.text.texwrap
+    import renpy.text.text
+    import renpy.text.extras
+
+    import renpy.gl
+    import renpy.gl2
+
+    import renpy.display.layout
+    import renpy.display.viewport
+    import renpy.display.transform
+    import renpy.display.motion # layout @UnresolvedImport
+    import renpy.display.behavior # layout @UnresolvedImport
+    import renpy.display.transition # core, layout @UnresolvedImport
+    import renpy.display.movetransition # core @UnresolvedImport
+    import renpy.display.im
+    import renpy.display.imagelike
+    import renpy.display.image # core, behavior, im, imagelike @UnresolvedImport
+    import renpy.display.video
+    import renpy.display.focus
+    import renpy.display.anim
+    import renpy.display.particle
+    import renpy.display.joystick
+    import renpy.display.controller
+    import renpy.display.minigame
+    import renpy.display.screen
+    import renpy.display.dragdrop
+    import renpy.display.imagemap
+    import renpy.display.predict
+    import renpy.display.emulator
+    import renpy.display.tts
+    import renpy.display.gesture
+    import renpy.display.matrix
+    import renpy.display.render
+
+    import renpy.display.error
+
+    # Note: For windows to work, renpy.audio.audio needs to be after
+    # renpy.display.module.
+
+    import renpy.audio
+    import renpy.audio.audio
+    import renpy.audio.music
+    import renpy.audio.sound
+
+    import renpy.ui
+    import renpy.screenlang
+
+    import renpy.sl2
+    import renpy.sl2.slast
+    import renpy.sl2.slparser
+    import renpy.sl2.slproperties
+    import renpy.sl2.sldisplayables
+
+    import renpy.lint
+    import renpy.warp
+
+    import renpy.editor
+
+    import renpy.memory
+
+    import renpy.exports
+    import renpy.character # depends on exports. @UnresolvedImport
+
+    import renpy.add_from
+    import renpy.dump
+
+    import renpy.minstore # depends on lots. @UnresolvedImport
+    import renpy.defaultstore # depends on everything. @UnresolvedImport
+
+    import renpy.test
+    import renpy.test.testmouse
+    import renpy.test.testfocus
+    import renpy.test.testkey
+    import renpy.test.testast
+    import renpy.test.testparser
+    import renpy.test.testexecution
+
+    import renpy.main
+
+    import renpy.defaultstore as store
 
     import renpy.arguments
 
+# This is here to help code analysis tools figure out all the cython modules that get imported.
+# noinspection PyUnreachableCode
+if False:
+
     import renpy.display.accelerator
-    import renpy.display.render
 
-    import renpy.gl.gl
-    import renpy.gl.gl1
     import renpy.gl.gldraw
-    import renpy.gl.glenviron_fixed
-    import renpy.gl.glenviron_limited
     import renpy.gl.glenviron_shader
     import renpy.gl.glrtt_copy
     import renpy.gl.glrtt_fbo
     import renpy.gl.gltexture
-
-    import renpy.angle.gl
-    import renpy.angle.gldraw
-    import renpy.angle.glenviron_shader
-    import renpy.angle.glrtt_copy
-    import renpy.angle.glrtt_fbo
-    import renpy.angle.gltexture
-
-
-if False:
-    import renpy.defaultstore as store
