@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -26,6 +26,9 @@
 
 # All functions in the is file should be documented in the wiki.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import sys
 import renpy.display
 import renpy.text
@@ -52,6 +55,9 @@ class Action(renpy.object.Object):
     def get_selected(self):
         return False
 
+    def get_tooltip(self):
+        return None
+
     def periodic(self, st):
         return
 
@@ -70,6 +76,8 @@ class BarValue(renpy.object.Object):
     # Alt text.
     alt = "Bar"
 
+    force_step = False
+
     def replaces(self, other):
         return
 
@@ -82,12 +90,15 @@ class BarValue(renpy.object.Object):
     def get_style(self):
         return "bar", "vbar"
 
+    def get_tooltip(self):
+        return None
 
 ##############################################################################
 # Things we can add to. These have two methods: add is called with the
 # widget we're adding. close is called when the thing is ready to be
 # closed.
 
+
 class Addable(object):
     # A style_prefix associates with this addable.
     style_prefix = None
@@ -210,6 +221,7 @@ class ChildOrFixed(Addable):
         if d is not None:
             raise Exception("Did not expect to close %r." % d)
 
+
 # A stack of things we can add to.
 stack = [ ]
 
@@ -234,10 +246,11 @@ def reset():
     at_stack = [ ]
     imagemap_stack = [ ]
 
+
 renpy.game.post_init.append(reset)
 
 
-def interact(type='misc', roll_forward=None, **kwargs):  # @ReservedAssignment
+def interact(type='misc', roll_forward=None, **kwargs): # @ReservedAssignment
     """
     :doc: ui
     :args: (roll_forward=None, mouse='default')
@@ -276,14 +289,13 @@ def interact(type='misc', roll_forward=None, **kwargs):  # @ReservedAssignment
         renpy.config.skipping = None
 
     if len(stack) != 1:
-        raise Exception("ui.interact called with non-empty widget/layer stack. Did you forget a ui.close() somewhere?\nStack was "+('\n'.join([str(item) for item in stack])))
+        raise Exception("ui.interact called with non-empty widget/layer stack. Did you forget a ui.close() somewhere?\nStack was " + ('\n'.join([str(item) for item in stack])))
 
     if at_stack:
         raise Exception("ui.interact called with non-empty at stack.")
 
     renpy.game.context().info._current_interact_type = type
     rv = renpy.game.interface.interact(roll_forward=roll_forward, **kwargs)
-    renpy.game.context().mark_seen()
     renpy.game.context().info._last_interact_type = type
 
     if renpy.exports.in_fixed_rollback() and roll_forward is not None:
@@ -377,7 +389,7 @@ def close(d=None):
 
 def reopen(w, clear):
 
-    stack.append(Many(w))
+    stack.append(Many(w, None, None))
 
     if clear:
         w.children[:] = [ ]
@@ -393,6 +405,7 @@ def context_enter(w):
 def context_exit(w):
     close(w)
 
+
 NoStylePrefixGiven = renpy.object.Sentinel("NoStylePrefixGiven")
 
 
@@ -407,7 +420,7 @@ def combine_style(style_prefix, style_suffix):
     else:
         new_style = style_prefix + "_" + style_suffix
 
-    return renpy.style.get_style(new_style)  # @UndefinedVariable
+    return renpy.style.get_style(new_style) # @UndefinedVariable
 
 
 def prefixed_style(style_suffix):
@@ -426,7 +439,10 @@ screen = None
 class Wrapper(renpy.object.Object):
 
     def __reduce__(self):
-        return self.name
+        if PY2:
+            return bytes(self.name)
+        else:
+            return self.name
 
     def __init__(self, function, one=False, many=False, imagemap=False, replaces=False, style=None, **kwargs):
 
@@ -460,7 +476,7 @@ class Wrapper(renpy.object.Object):
 
         # Pull out the special kwargs, widget_id, at, and style_prefix.
 
-        widget_id = kwargs.pop("id", None)  # @ReservedAssignment
+        widget_id = kwargs.pop("id", None) # @ReservedAssignment
 
         at_list = kwargs.pop("at", [ ])
         if not isinstance(at_list, (list, tuple)):
@@ -513,9 +529,9 @@ class Wrapper(renpy.object.Object):
             etype, e, tb = sys.exc_info(); etype
 
             if tb.tb_next is None:
-                e.args = (e.args[0].replace("__call__", "ui." + self.name), )
+                e.args = (e.args[0].replace("__call__", "ui." + self.name),)
 
-            del tb  # Important! Prevents memory leaks via our frame.
+            del tb # Important! Prevents memory leaks via our frame.
             raise
 
         main = w._main or w
@@ -567,10 +583,10 @@ class Wrapper(renpy.object.Object):
 
         return main
 
-
 ##############################################################################
 # Widget functions.
 
+
 def _add(d, **kwargs):
     d = renpy.easy.displayable(d)
 
@@ -585,6 +601,7 @@ def _add(d, **kwargs):
 
     return rv
 
+
 add = Wrapper(_add)
 
 
@@ -596,6 +613,7 @@ def _implicit_add(d):
 
     return d
 
+
 implicit_add = Wrapper(_implicit_add)
 
 
@@ -608,6 +626,7 @@ def _image(im, **properties):
 
     return d
 
+
 image = Wrapper(_image)
 
 null = Wrapper(renpy.display.layout.Null)
@@ -615,6 +634,7 @@ text = Wrapper(renpy.text.text.Text, style="text", replaces=True)
 hbox = Wrapper(renpy.display.layout.MultiBox, layout="horizontal", style="hbox", many=True)
 vbox = Wrapper(renpy.display.layout.MultiBox, layout="vertical", style="vbox", many=True)
 fixed = Wrapper(renpy.display.layout.MultiBox, layout="fixed", style="fixed", many=True)
+default_fixed = Wrapper(renpy.display.layout.MultiBox, layout="fixed", many=True)
 grid = Wrapper(renpy.display.layout.Grid, style="grid", many=True)
 side = Wrapper(renpy.display.layout.Side, style="side", many=True)
 
@@ -622,6 +642,7 @@ side = Wrapper(renpy.display.layout.Side, style="side", many=True)
 def _sizer(maxwidth=None, maxheight=None, **properties):
     return renpy.display.layout.Container(xmaximum=maxwidth, ymaximum=maxheight, **properties)
 
+
 sizer = Wrapper(_sizer, one=True)
 window = Wrapper(renpy.display.layout.Window, style="window", one=True, child=None)
 frame = Wrapper(renpy.display.layout.Window, style="frame", one=True, child=None)
@@ -634,10 +655,13 @@ soundstopbehavior = Wrapper(renpy.display.behavior.SoundStopBehavior)
 
 def _key(key, action=None, activate_sound=None):
 
-    if action is None:
-        raise Exception("Action is required in ui.key.")
+    if isinstance(key, (list, tuple)):
+        keymap = {k: action for k in key}
+    else:
+        keymap = {key: action}
+
+    return renpy.display.behavior.Keymap(activate_sound=activate_sound, **keymap)
 
-    return renpy.display.behavior.Keymap(activate_sound=activate_sound, **{ key : action})
 
 key = Wrapper(_key)
 
@@ -652,10 +676,13 @@ class ChoiceActionBase(Action):
     previously visited and mark it so if it is chosen.
     """
 
-    def __init__(self, label, value, location=None, block_all=None):
+    sensitive = True
+
+    def __init__(self, label, value, location=None, block_all=None, sensitive=True, args=None, kwargs=None):
         self.label = label
         self.value = value
         self.location = location
+        self.sensitive = sensitive
 
         if block_all is None:
             self.block_all = renpy.config.fix_rollback_without_choice
@@ -665,13 +692,18 @@ class ChoiceActionBase(Action):
         self.chosen = None
 
         if self.location:
-            self.chosen = renpy.game.persistent._chosen  # @UndefinedVariable
+            self.chosen = renpy.game.persistent._chosen # @UndefinedVariable
 
             if self.chosen is None:
                 self.chosen = renpy.game.persistent._chosen = { }
 
+        # The arguments passed to a menu choice.
+        self.args = args
+        self.kwargs = kwargs
+
     def get_sensitive(self):
-        return not renpy.exports.in_fixed_rollback() or (not self.block_all and self.get_selected())
+        return (self.sensitive and
+                not renpy.exports.in_fixed_rollback() or (not self.block_all and self.get_selected()))
 
     def get_selected(self):
         roll_forward = renpy.exports.roll_forward_info()
@@ -793,8 +825,6 @@ def menu(menuitems,
          default=False,
          **properties):
 
-    # menu is now a conglomeration of other widgets. And bully for it.
-
     renpy.ui.vbox(style=style, **properties)
 
     for label, val in menuitems:
@@ -805,7 +835,10 @@ def menu(menuitems,
             text = choice_style
             button = choice_button_style
 
-            clicked = ChoiceReturn(label, val, location)
+            if isinstance(val, ChoiceReturn):
+                clicked = val
+            else:
+                clicked = ChoiceReturn(label, val, location)
 
             if clicked.get_chosen():
                 text = choice_chosen_style
@@ -828,7 +861,8 @@ def menu(menuitems,
 
     close()
 
-input = Wrapper(renpy.display.behavior.Input, exclude='{}', style="input", replaces=True)  # @ReservedAssignment
+
+input = Wrapper(renpy.display.behavior.Input, exclude='{}', style="input", replaces=True) # @ReservedAssignment
 
 
 def imagemap_compat(ground,
@@ -873,6 +907,7 @@ def imagemap_compat(ground,
 
     close()
 
+
 button = Wrapper(renpy.display.behavior.Button, style='button', one=True)
 
 
@@ -902,10 +937,15 @@ def _imagebutton(idle_image=None,
             return b
 
         if auto is not None:
-            return renpy.config.imagemap_auto_function(auto, name)
+            rv = renpy.config.imagemap_auto_function(auto, name)
+            if rv is not None:
+                return rv
 
         if required:
-            raise Exception("Could not find a %s image for imagemap." % name)
+            if auto:
+                raise Exception("Imagebutton does not have a %s image. (auto=%r)." % (name, auto))
+            else:
+                raise Exception("Imagebutton does not have a %s image." % (name,))
 
         return None
 
@@ -927,6 +967,7 @@ def _imagebutton(idle_image=None,
         selected_activate_image=selected_activate_image,
         **properties)
 
+
 imagebutton = Wrapper(_imagebutton, style="image_button")
 
 
@@ -945,7 +986,7 @@ def _textbutton(label, clicked=None, style=None, text_style=None, substitute=Tru
         style = prefixed_style("button")
 
     if text_style is None:
-        text_style = renpy.style.get_text_style(style, prefixed_style('button_text'))  # @UndefinedVariable
+        text_style = renpy.style.get_text_style(style, prefixed_style('button_text')) # @UndefinedVariable
 
     rv = renpy.display.behavior.Button(style=style, clicked=clicked, **button_kwargs)
     text = renpy.text.text.Text(label, style=text_style, substitute=substitute, scope=scope, **text_kwargs)
@@ -954,6 +995,7 @@ def _textbutton(label, clicked=None, style=None, text_style=None, substitute=Tru
     rv._composite_parts = [ text ]
     return rv
 
+
 textbutton = Wrapper(_textbutton)
 
 
@@ -965,7 +1007,7 @@ def _label(label, style=None, text_style=None, substitute=True, scope=None, **kw
         style = prefixed_style('label')
 
     if text_style is None:
-        text_style = renpy.style.get_text_style(style, prefixed_style('label_text'))  # @UndefinedVariable
+        text_style = renpy.style.get_text_style(style, prefixed_style('label_text')) # @UndefinedVariable
 
     rv = renpy.display.layout.Window(None, style=style, **label_kwargs)
     text = renpy.text.text.Text(label, style=text_style, substitute=substitute, scope=scope, **text_kwargs)
@@ -974,6 +1016,7 @@ def _label(label, style=None, text_style=None, substitute=True, scope=None, **kw
     rv._composite_parts = [ text ]
     return rv
 
+
 label = Wrapper(_label)
 
 adjustment = renpy.display.behavior.Adjustment
@@ -982,13 +1025,13 @@ adjustment = renpy.display.behavior.Adjustment
 def _bar(*args, **properties):
 
     if len(args) == 4:
-        width, height, range, value = args  # @ReservedAssignment
+        width, height, range, value = args # @ReservedAssignment
     if len(args) == 2:
-        range, value = args  # @ReservedAssignment
+        range, value = args # @ReservedAssignment
         width = None
         height = None
     else:
-        range = 1  # @ReservedAssignment
+        range = 1 # @ReservedAssignment
         value = 0
         width = None
         height = None
@@ -997,10 +1040,10 @@ def _bar(*args, **properties):
         width = properties.pop("width")
 
     if "height" in properties:
-        height  = properties.pop("height")
+        height = properties.pop("height")
 
     if "range" in properties:
-        range = properties.pop("range")  # @ReservedAssignment
+        range = properties.pop("range") # @ReservedAssignment
 
     if "value" in properties:
         value = properties.pop("value")
@@ -1019,6 +1062,7 @@ def _bar(*args, **properties):
 
     return renpy.display.behavior.Bar(range, value, width, height, **properties)
 
+
 bar = Wrapper(_bar, vertical=False, replaces=True)
 vbar = Wrapper(_bar, vertical=True, replaces=True)
 slider = Wrapper(_bar, style='slider', replaces=True)
@@ -1027,7 +1071,7 @@ scrollbar = Wrapper(_bar, style='scrollbar', replaces=True)
 vscrollbar = Wrapper(_bar, style='vscrollbar', replaces=True)
 
 
-def _autobar_interpolate(range, start, end, time, st, at, **properties):  # @ReservedAssignment
+def _autobar_interpolate(range, start, end, time, st, at, **properties): # @ReservedAssignment
 
     if st > time:
         t = 1.0
@@ -1039,42 +1083,70 @@ def _autobar_interpolate(range, start, end, time, st, at, **properties):  # @Res
     value = start + t * (end - start)
     return renpy.display.behavior.Bar(range, value, None, None, **properties), redraw
 
+
 autobar_interpolate = renpy.curry.curry(_autobar_interpolate)
 
 
-def _autobar(range, start, end, time, **properties):  # @ReservedAssignment
+def _autobar(range, start, end, time, **properties): # @ReservedAssignment
     return renpy.display.layout.DynamicDisplayable(autobar_interpolate(range, start, end, time, **properties))
 
+
 autobar = Wrapper(_autobar)
 transform = Wrapper(renpy.display.motion.Transform, one=True, style='transform')
 
 _viewport = Wrapper(renpy.display.viewport.Viewport, one=True, replaces=True, style='viewport')
 _vpgrid = Wrapper(renpy.display.viewport.VPGrid, many=True, replaces=True, style='vpgrid')
 
+VIEWPORT_SIZE = 32767
 
-def viewport_common(vpfunc, scrollbars=None, **properties):
+
+def viewport_common(vpfunc, _spacing_to_side, scrollbars=None, **properties):
 
     if scrollbars is None:
         return vpfunc(**properties)
 
-    viewport_properties = { }
-    side_properties = { }
+    (vscrollbar_properties, scrollbar_properties, side_properties, viewport_properties, core_properties) = \
+        renpy.easy.split_properties(properties, "vscrollbar_", "scrollbar_", "side_", "viewport_", "")
 
-    for k, v in properties.iteritems():
-        if k.startswith("side_"):
-            side_properties[k[5:]] = v
-        else:
-            viewport_properties[k] = v
+    if renpy.config.position_viewport_side:
+        from renpy.sl2.slproperties import position_property_names
+
+        for k, v in core_properties.items():
+            if k in position_property_names:
+                side_properties[k] = v
+            elif _spacing_to_side and (k == "spacing"):
+                side_properties[k] = v
+            else:
+                viewport_properties[k] = v
+
+    else:
+        viewport_properties.update(core_properties)
+
+    if renpy.config.prefix_viewport_scrollbar_styles and (scrollbars != "vertical"):
+        scrollbar_properties.setdefault("style", prefixed_style("scrollbar"))
+    else:
+        scrollbar_properties.setdefault("style", "scrollbar")
+
+    if renpy.config.prefix_viewport_scrollbar_styles and (scrollbars != "horizontal"):
+        vscrollbar_properties.setdefault("style", prefixed_style("vscrollbar"))
+    else:
+        vscrollbar_properties.setdefault("style", "vscrollbar")
 
     alt = viewport_properties.get("alt", "viewport")
+    scrollbar_properties.setdefault("alt", renpy.minstore.__(alt) + " " + renpy.minstore.__("horizontal scroll"))
+    vscrollbar_properties.setdefault("alt", renpy.minstore.__(alt) + " " + renpy.minstore.__("vertical scroll"))
 
     if scrollbars == "vertical":
+
+        if renpy.config.scrollbar_child_size:
+            viewport_properties.setdefault("child_size", (None, VIEWPORT_SIZE))
+
         side("c r", **side_properties)
 
         rv = vpfunc(**viewport_properties)
         addable = stack.pop()
 
-        vscrollbar(adjustment=rv.yadjustment, alt=alt + " vertical scrollbar")
+        vscrollbar(adjustment=rv.yadjustment, **vscrollbar_properties)
         close()
 
         stack.append(addable)
@@ -1082,12 +1154,16 @@ def viewport_common(vpfunc, scrollbars=None, **properties):
         return rv
 
     elif scrollbars == "horizontal":
+
+        if renpy.config.scrollbar_child_size:
+            viewport_properties.setdefault("child_size", (VIEWPORT_SIZE, None))
+
         side("c b", **side_properties)
 
         rv = vpfunc(**viewport_properties)
         addable = stack.pop()
 
-        scrollbar(adjustment=rv.xadjustment, alt=alt + " horizontal scrollbar")
+        scrollbar(adjustment=rv.xadjustment, **scrollbar_properties)
         close()
 
         stack.append(addable)
@@ -1096,13 +1172,16 @@ def viewport_common(vpfunc, scrollbars=None, **properties):
 
     else:
 
+        if renpy.config.scrollbar_child_size:
+            viewport_properties.setdefault("child_size", (VIEWPORT_SIZE, VIEWPORT_SIZE))
+
         side("c r b", **side_properties)
 
         rv = vpfunc(**viewport_properties)
         addable = stack.pop()
 
-        vscrollbar(adjustment=rv.yadjustment, alt=alt + " vertical scrollbar")
-        scrollbar(adjustment=rv.xadjustment, alt=alt +" horizontal scrollbar")
+        vscrollbar(adjustment=rv.yadjustment, **vscrollbar_properties)
+        scrollbar(adjustment=rv.xadjustment, **scrollbar_properties)
         close()
 
         stack.append(addable)
@@ -1111,11 +1190,12 @@ def viewport_common(vpfunc, scrollbars=None, **properties):
 
 
 def viewport(**properties):
-    return viewport_common(_viewport, **properties)
+    return viewport_common(_viewport, True, **properties)
 
 
 def vpgrid(**properties):
-    return viewport_common(_vpgrid, **properties)
+    return viewport_common(_vpgrid, False, **properties)
+
 
 conditional = Wrapper(renpy.display.behavior.Conditional, one=True)
 timer = Wrapper(renpy.display.behavior.Timer, replaces=True)
@@ -1123,10 +1203,10 @@ drag = Wrapper(renpy.display.dragdrop.Drag, replaces=True, one=True)
 draggroup = Wrapper(renpy.display.dragdrop.DragGroup, replaces=True, many=True)
 mousearea = Wrapper(renpy.display.behavior.MouseArea, replaces=True)
 
-
 ##############################################################################
 # New-style imagemap related functions.
 
+
 class Imagemap(object):
     """
     Stores information about the images used by an imagemap.
@@ -1159,22 +1239,23 @@ def _imagemap(ground=None, hover=None, insensitive=None, idle=None, selected_hov
             return variable
 
         if auto:
-            fn = renpy.config.imagemap_auto_function(auto, name)
-            if fn is not None:
-                return fn
+            for i in name:
+                fn = renpy.config.imagemap_auto_function(auto, i)
+                if fn is not None:
+                    return fn
 
         if other is not None:
             return other
 
-        raise Exception("Could not find a %s image for imagemap." % name)
+        raise Exception("Could not find a %s image for imagemap." % name[0])
 
-    ground = pick(ground, "ground", None)
-    idle = pick(idle, "idle", ground)
-    selected_idle = pick(selected_idle, "selected_idle", idle)
-    hover = pick(hover, "hover", ground)
-    selected_hover = pick(selected_hover, "selected_hover", hover)
-    insensitive = pick(insensitive, "insensitive", ground)
-    selected_insensitive = pick(selected_insensitive, "selected_insensitive", hover)
+    ground = pick(ground, ("ground", "idle"), idle)
+    idle = pick(idle, ("idle",), ground)
+    selected_idle = pick(selected_idle, ("selected_idle",), idle)
+    hover = pick(hover, ("hover",), ground)
+    selected_hover = pick(selected_hover, ("selected_hover",), hover)
+    insensitive = pick(insensitive, ("insensitive",), ground)
+    selected_insensitive = pick(selected_insensitive, ("selected_insensitive",), hover)
 
     imagemap_stack.append(
         Imagemap(
@@ -1205,6 +1286,7 @@ def _imagemap(ground=None, hover=None, insensitive=None, idle=None, selected_hov
 
     return rv
 
+
 imagemap = Wrapper(_imagemap, imagemap=True, style='imagemap')
 
 
@@ -1258,6 +1340,7 @@ def _hotspot(spot, style='hotspot', **properties):
         style=style,
         **properties)
 
+
 hotspot_with_child = Wrapper(_hotspot, style="hotspot", one=True)
 
 
@@ -1266,7 +1349,7 @@ def hotspot(*args, **kwargs):
     null()
 
 
-def _hotbar(spot, adjustment=None, range=None, value=None, **properties):  # @ReservedAssignment
+def _hotbar(spot, adjustment=None, range=None, value=None, **properties): # @ReservedAssignment
 
     if (adjustment is None) and (range is None) and (value is None):
         raise Exception("hotbar requires either an adjustment or a range and value.")
@@ -1283,10 +1366,10 @@ def _hotbar(spot, adjustment=None, range=None, value=None, **properties):  # @Re
     properties.setdefault("xanchor", 0)
     properties.setdefault("yanchor", 0)
 
-    fore_bar=imagemap.cache.crop(imagemap.selected_idle, spot)
-    aft_bar=imagemap.cache.crop(imagemap.idle, spot)
-    hover_fore_bar=imagemap.cache.crop(imagemap.selected_hover, spot)
-    hover_aft_bar=imagemap.cache.crop(imagemap.hover, spot)
+    fore_bar = imagemap.cache.crop(imagemap.selected_idle, spot)
+    aft_bar = imagemap.cache.crop(imagemap.idle, spot)
+    hover_fore_bar = imagemap.cache.crop(imagemap.selected_hover, spot)
+    hover_aft_bar = imagemap.cache.crop(imagemap.hover, spot)
 
     if h > w:
         properties.setdefault("bar_vertical", True)
@@ -1313,16 +1396,18 @@ def _hotbar(spot, adjustment=None, range=None, value=None, **properties):  # @Re
         ymaximum=h,
         **properties)
 
-hotbar = Wrapper(_hotbar, style="hotbar", replaces=True)
 
+hotbar = Wrapper(_hotbar, style="hotbar", replaces=True)
 
 ##############################################################################
 # Curried functions, for use in clicked, hovered, and unhovered.
 
+
 def _returns(v):
 
     return v
 
+
 returns = renpy.curry.curry(_returns)
 
 
@@ -1336,6 +1421,7 @@ def _jumps(label, transition=None):
 
     raise renpy.exports.jump(label)
 
+
 jumps = renpy.curry.curry(_jumps)
 
 
@@ -1343,6 +1429,7 @@ def _jumpsoutofcontext(label):
 
     raise renpy.game.JumpOutException(label)
 
+
 jumpsoutofcontext = renpy.curry.curry(_jumpsoutofcontext)
 
 
@@ -1364,6 +1451,7 @@ def gamemenus(*args):
 ##############################################################################
 # The on statement.
 
+
 on = Wrapper(renpy.display.behavior.OnEvent)
 
 ##############################################################################
@@ -1386,8 +1474,9 @@ def screen_id(id_, d):
 ##############################################################################
 # Postamble
 
+
 # Update the wrappers to have names.
 k, v = None, None
-for k, v in globals().iteritems():
+for k, v in globals().items():
     if isinstance(v, Wrapper):
         v.name = k
